@@ -1,11 +1,14 @@
+/*
+ * $PostgreSQL: pgsql/src/backend/port/nextstep/port.c,v 1.12 2009/06/11 14:49:00 momjian Exp $
+ */
+#include "postgres.h"
+
 #ifndef _POSIX_SOURCE
 #include "libc.h"
 #else
 #include <unistd.h>
-#include <stdlib.h>
 #endif
 
-#include <string.h>
 #include <sys/signal.h>
 
 
